/* $Id: ssp_pf2.c,v 3.1 2009/04/14 18:50:45 fknobbe Exp $
 *
 * Copyright (c) 2003 Hector Paterno <apaterno@dsnsecurity.com>
 * Copyright (c) 2004, 2005 Olaf Schreck <chakl@syscall.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * ssp_pf2.c
 *
 * Purpose:
 *
 * This SnortSam plugin is meant for dynamic (un)blocking on the PF firewall
 * for OpenBSD >= 3.5, FreeBSD and NetBSD.
 * SnortSam will expire the blocks itself since PF does not have automatic 
 * time-out functionality.
 *
 * This is a reimplementation of the original ssp_pf plugin in order to 
 * simplify it and make it portable.
 */

#if defined(OpenBSD) || defined(FreeBSD) || defined(NetBSD)

#ifndef		__SSP_PF2_C__
#define		__SSP_PF2_C__

#include "snortsam.h"
#include "ssp_pf2.h"


/* Routine for opt parsing ( opt=value opt2=value2 etc. ) */
int parse_opts(char *line, opt_pf2 *opt, char *sep, char *int_sep, int nopt)
{
   char *last;
   char *last2;
   char *pt;
   char *pt2;
   int fo=0;
   int di=0;
   
   if((line==NULL) || (opt==NULL) || (sep==NULL) || (int_sep==NULL))
     return -1;

   for(pt=strtok_r(line, sep, &last); pt; pt=strtok_r(NULL, sep, &last))
     {
	for(pt2=strtok_r(pt, int_sep, &last2), fo=0; pt2; pt2=strtok_r(NULL, int_sep, &last2))
	  {	     
	     if(fo==0)
	       {
		  for(; fo<nopt; fo++)
		    {
		      if (strncmp(opt[fo].name, pt2, MAX_OPT_NAME)==0)
			 {
			    fo++;
			    di++;
			    break;
			 }		       
		    }		  
	       }
	     else
	       {
		  if(di)
		    if(opt[--fo].vt>0)
		      strncpy(opt[fo].v.value_s, pt2, MAX_OPT_VALUE);
		    else
		      opt[fo].v.value_d=atoi(pt2);
		  fo=0;
		  di=0;
	       }	     
	  }
     }
   
   return;
}


/*
 * This routine parses the pf statements in the config file.
 */
void
PF2Parse(char *val, char *file, unsigned long line, DATALIST * plugindatalist)
{
   PF2DATA        *pfp = NULL;
   char           msg[STRBUFSIZE + 2];
   char           tbuf[PF_TABLE_NAME_SIZE];
   opt_pf2        options[2]={	
	{"anchor", "", 1},
	{"table",  "", 1}
   };

#ifdef FWSAMDEBUG
   printf("Debug: [pf2] Plugin Parsing...\n");
#endif
   
   if (*val)
     {
	
	if(parse_opts(val, options, " \t", "=", (sizeof(options)/sizeof(opt_pf2)))<0)
	  {
	     snprintf(msg, sizeof(msg) - 1, "Error: [%s: %lu] invalid PF parameters !. PF2 Plugin disabled.", file, line);
	     logmessage(1, msg, "pf2", 0);
	     plugindatalist->data=NULL;			
	     return;
	  }

	pfp = safemalloc(sizeof(PF2DATA), "PF2Parse", "pfp");
	bzero(pfp, sizeof(PF2DATA));
	plugindatalist->data = pfp;

	/* Check Anchor */
	if(strlen(options[PF2_OPT_ANCHOR].v.value_s)<1)
	  {
	     snprintf(msg, sizeof(msg) - 1, 
		      "Info: [%s: %lu] PF anchor name not defined, using \"snortsam\"", file, line);
	     logmessage(1, msg, "pf2", 0);
	     safecopy(pfp->anchorname, "snortsam");	/* save anchorname */
	  }
	else
	  {
	     safecopy(pfp->anchorname, options[PF2_OPT_ANCHOR].v.value_s);	/* save anchorname */		
	  }
	
	/* Check Table */
	
	if(strlen(options[PF2_OPT_TABLE].v.value_s)<1)
	  {
	     snprintf(msg, sizeof(msg) - 1, 
		      "Info: [%s: %lu] PF table name not defined, using \"blockin,blockout\"", file, line);
	     logmessage(1, msg, "pf2", 0);
	     safecopy(pfp->tablein,  "blockin");
	     safecopy(pfp->tableout, "blockout");
	  }
	else
	  {
	    /* save tablenames */	     
	    snprintf(tbuf, PF_TABLE_NAME_SIZE, "%s_in", 
		     options[PF2_OPT_TABLE].v.value_s);
	    safecopy(pfp->tablein,  tbuf);
	    snprintf(tbuf, PF_TABLE_NAME_SIZE, "%s_out", 
		     options[PF2_OPT_TABLE].v.value_s);
	    safecopy(pfp->tableout, tbuf);
	  }

     }

#ifdef FWSAMDEBUG
   printf("Debug: [pf2] Adding PF: \n");
   printf("anchor=%s\ntables=%s,%s\n",
	  pfp->anchorname, pfp->tablein, pfp->tableout);
#endif

}


/*
 * BLOCK/UNBLOCK Routine
 */
void
PF2Block(BLOCKINFO * bd, void * data,unsigned long qp) 
{
	PF2DATA	*pfp;
	int	 pfdev;
	int	 tin=0, tout=0;
	char	 ipsrc[256];		/* ip as a string */
	char	 msg[STRBUFSIZE + 2];
#ifdef FWSAMDEBUG
	pthread_t threadid=pthread_self();
#endif
   
	if(!data)
		return;
   
	pfp=(PF2DATA *)data;
   
#ifdef FWSAMDEBUG
	printf("Debug: [pf][%lx] Plugin Blocking...\n", threadid);
#endif

	snprintf(ipsrc, sizeof(ipsrc) - 1, inettoa(bd->blockip));
	switch(bd->mode & FWSAM_HOW)
	{	
	case FWSAM_HOW_THIS:
		tin = tout = 1;
		break;
	case FWSAM_HOW_IN:
		tin = 1;
		break;
	case FWSAM_HOW_OUT:
		tout = 1;
		break;
	case FWSAM_HOW_INOUT:
		tin = tout = 1;
		break;
	}

	/* open the pf device */
	pfdev = open(PFDEV, O_RDWR);
	if (pfdev == -1) {
		logmessage(1, "cannot open packet filter device", "pf2", 0);
	}

	/* BLOCK */
	if (bd->block)
	{
		snprintf(msg, sizeof(msg) - 1, "Info: Blocking ip %s", ipsrc);
		logmessage(3, msg, "pf2", 0);

		if (tin) 
			change_table(pfdev, 1, pfp->tablein, 
				     pfp->anchorname, ipsrc);
		if (tout)
			change_table(pfdev, 1, pfp->tableout, 
				     pfp->anchorname, ipsrc);
	}
	else   /* UNBLOCK */
	{
		snprintf(msg, sizeof(msg) - 1, "Info: Unblocking ip %s", ipsrc);
		logmessage(3, msg, "pf2", 0);

		if (tin)
			change_table(pfdev, 0, pfp->tablein, 
				     pfp->anchorname, ipsrc);
		if (tout)
			change_table(pfdev, 0, pfp->tableout, 
				     pfp->anchorname, ipsrc);
	}
	close(pfdev);
	return;
}				       /* PF2BLOCK */

/* borrowed from OpenBSDs pfctl code */
int
change_table(int pfdev, int add, const char *table, const char *anchor, const char *ipsrc)
{
	char   msg[STRBUFSIZE + 2];
	struct pfioc_table	io;
	struct pfr_addr		addr;

	bzero(&io, sizeof(io));
	strlcpy(io.pfrio_table.pfrt_name, table, sizeof(io.pfrio_table.pfrt_name));
	strlcpy(io.pfrio_table.pfrt_anchor, anchor, sizeof(io.pfrio_table.pfrt_anchor));
	io.pfrio_buffer = &addr;
	io.pfrio_esize = sizeof(addr);
	io.pfrio_size = 1;

	bzero(&addr, sizeof(addr));
	if (ipsrc == NULL || !ipsrc[0])
		return (-1);
	if (inet_pton(AF_INET, ipsrc, &addr.pfra_ip4addr) == 1) {
		addr.pfra_af = AF_INET;
		addr.pfra_net = 32;
	} else if (inet_pton(AF_INET6, ipsrc, &addr.pfra_ip6addr) == 1) {
		addr.pfra_af = AF_INET6;
		addr.pfra_net = 128;
	} else {
	        snprintf(msg, sizeof(msg) - 1, "invalid ipsrc");
                logmessage(3, msg, "pf2", 0);
		return (-1);
	}

	if (ioctl(pfdev, add ? DIOCRADDADDRS : DIOCRDELADDRS, &io) &&
	    errno != ESRCH) {
	        snprintf(msg, sizeof(msg) - 1, 
			 "cannot %s %s %s table %s: %s",
			 add ? "add" : "remove", ipsrc, add ? "to" : "from", 
			 table, strerror(errno));
                logmessage(3, msg, "pf2", 0);
		return (-1);
	}
	return (0);
}

#endif				/* __SSP_PF2_C__ */

#endif /* OpenBSD || FreeBSD || NetBSD */
